package org.apache.bcel.verifier.structurals;

import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import org.apache.bcel.generic.Visitor;
import java.util.ArrayList;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.Select;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.ATHROW;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.RET;
import java.util.List;
import java.util.Arrays;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import java.util.HashMap;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.InstructionHandle;
import java.util.Map;

public class ControlFlowGraph
{
    private final Subroutines subroutines;
    private final ExceptionHandlers exceptionhandlers;
    private final Map<InstructionHandle, InstructionContext> instructionContexts;
    
    public ControlFlowGraph(final MethodGen methodGen) {
        this(methodGen, true);
    }
    
    public ControlFlowGraph(final MethodGen methodGen, final boolean enableJustIceCheck) {
        this.instructionContexts = new HashMap<InstructionHandle, InstructionContext>();
        this.subroutines = new Subroutines(methodGen, enableJustIceCheck);
        this.exceptionhandlers = new ExceptionHandlers(methodGen);
        final InstructionHandle[] instructionHandles;
        final InstructionHandle[] instructionhandles = instructionHandles = methodGen.getInstructionList().getInstructionHandles();
        for (final InstructionHandle instructionhandle : instructionHandles) {
            this.instructionContexts.put(instructionhandle, new InstructionContextImpl(instructionhandle));
        }
    }
    
    public InstructionContext contextOf(final InstructionHandle inst) {
        final InstructionContext ic = this.instructionContexts.get(inst);
        if (ic == null) {
            throw new AssertionViolatedException("InstructionContext requested for an InstructionHandle that's not known!");
        }
        return ic;
    }
    
    public InstructionContext[] contextsOf(final InstructionHandle[] insts) {
        final InstructionContext[] ret = new InstructionContext[insts.length];
        Arrays.setAll(ret, i -> this.contextOf(insts[i]));
        return ret;
    }
    
    public InstructionContext[] getInstructionContexts() {
        final InstructionContext[] ret = new InstructionContext[this.instructionContexts.size()];
        return this.instructionContexts.values().toArray(ret);
    }
    
    public boolean isDead(final InstructionHandle i) {
        return this.subroutines.subroutineOf(i) == null;
    }
    
    private class InstructionContextImpl implements InstructionContext
    {
        private int TAG;
        private final InstructionHandle instruction;
        private final Map<InstructionContext, Frame> inFrames;
        private final Map<InstructionContext, Frame> outFrames;
        private List<InstructionContext> executionPredecessors;
        
        public InstructionContextImpl(final InstructionHandle inst) {
            if (inst == null) {
                throw new AssertionViolatedException("Cannot instantiate InstructionContextImpl from NULL.");
            }
            this.instruction = inst;
            this.inFrames = new HashMap<InstructionContext, Frame>();
            this.outFrames = new HashMap<InstructionContext, Frame>();
        }
        
        private InstructionHandle[] _getSuccessors() {
            final InstructionHandle[] single = { null };
            final Instruction inst = this.getInstruction().getInstruction();
            if (inst instanceof RET) {
                final Subroutine s = ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction());
                if (s == null) {
                    throw new AssertionViolatedException("Asking for successors of a RET in dead code?!");
                }
                final InstructionHandle[] jsrs = s.getEnteringJsrInstructions();
                final InstructionHandle[] ret = new InstructionHandle[jsrs.length];
                Arrays.setAll(ret, i -> jsrs[i].getNext());
                return ret;
            }
            else {
                if (inst instanceof ReturnInstruction || inst instanceof ATHROW) {
                    return InstructionHandle.EMPTY_ARRAY;
                }
                if (inst instanceof JsrInstruction) {
                    single[0] = ((JsrInstruction)inst).getTarget();
                    return single;
                }
                if (inst instanceof GotoInstruction) {
                    single[0] = ((GotoInstruction)inst).getTarget();
                    return single;
                }
                if (!(inst instanceof BranchInstruction)) {
                    single[0] = this.getInstruction().getNext();
                    return single;
                }
                if (inst instanceof Select) {
                    final InstructionHandle[] matchTargets = ((Select)inst).getTargets();
                    final InstructionHandle[] ret2 = new InstructionHandle[matchTargets.length + 1];
                    ret2[0] = ((Select)inst).getTarget();
                    System.arraycopy(matchTargets, 0, ret2, 1, matchTargets.length);
                    return ret2;
                }
                final InstructionHandle[] pair = { this.getInstruction().getNext(), ((BranchInstruction)inst).getTarget() };
                return pair;
            }
        }
        
        @Override
        public boolean execute(final Frame inFrame, final ArrayList<InstructionContext> execPreds, final InstConstraintVisitor icv, final ExecutionVisitor ev) {
            final List<InstructionContext> clone = (List<InstructionContext>)execPreds.clone();
            this.executionPredecessors = clone;
            if ((this.lastExecutionJSR() == null && ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction()) != ControlFlowGraph.this.subroutines.getTopLevel()) || (this.lastExecutionJSR() != null && ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction()) == ControlFlowGraph.this.subroutines.getTopLevel())) {
                throw new AssertionViolatedException("Huh?! Am I '" + this + "' part of a subroutine or not?");
            }
            Frame inF = this.inFrames.get(this.lastExecutionJSR());
            if (inF == null) {
                this.inFrames.put(this.lastExecutionJSR(), inFrame);
                inF = inFrame;
            }
            else if (inF.equals(inFrame) || !this.mergeInFrames(inFrame)) {
                return false;
            }
            final Frame workingFrame = inF.getClone();
            try {
                icv.setFrame(workingFrame);
                this.getInstruction().accept(icv);
            }
            catch (final StructuralCodeConstraintException ce) {
                ce.extendMessage("", "\nInstructionHandle: " + this.getInstruction() + "\n");
                ce.extendMessage("", "\nExecution Frame:\n" + workingFrame);
                this.extendMessageWithFlow(ce);
                throw ce;
            }
            ev.setFrame(workingFrame);
            this.getInstruction().accept(ev);
            this.outFrames.put(this.lastExecutionJSR(), workingFrame);
            return true;
        }
        
        private void extendMessageWithFlow(final StructuralCodeConstraintException e) {
            final String s = "Execution flow:\n";
            e.extendMessage("", "Execution flow:\n" + this.getExecutionChain());
        }
        
        @Override
        public ExceptionHandler[] getExceptionHandlers() {
            return ControlFlowGraph.this.exceptionhandlers.getExceptionHandlers(this.getInstruction());
        }
        
        private String getExecutionChain() {
            final StringBuilder s = new StringBuilder(this.toString());
            for (int i = this.executionPredecessors.size() - 1; i >= 0; --i) {
                s.insert(0, this.executionPredecessors.get(i) + "\n");
            }
            return s.toString();
        }
        
        @Override
        public Frame getInFrame() {
            final InstructionContext jsr = this.lastExecutionJSR();
            final Frame org = this.inFrames.get(jsr);
            if (org == null) {
                throw new AssertionViolatedException("inFrame not set! This:\n" + this + "\nInFrames: '" + this.inFrames + "'.");
            }
            return org.getClone();
        }
        
        @Override
        public InstructionHandle getInstruction() {
            return this.instruction;
        }
        
        @Override
        public Frame getOutFrame(final ArrayList<InstructionContext> execChain) {
            this.executionPredecessors = execChain;
            final InstructionContext jsr = this.lastExecutionJSR();
            final Frame org = this.outFrames.get(jsr);
            if (org == null) {
                throw new AssertionViolatedException("outFrame not set! This:\n" + this + "\nExecutionChain: " + this.getExecutionChain() + "\nOutFrames: '" + this.outFrames + "'.");
            }
            return org.getClone();
        }
        
        @Override
        public InstructionContext[] getSuccessors() {
            return ControlFlowGraph.this.contextsOf(this._getSuccessors());
        }
        
        @Override
        public int getTag() {
            return this.TAG;
        }
        
        private InstructionContextImpl lastExecutionJSR() {
            final int size = this.executionPredecessors.size();
            int retcount = 0;
            for (int i = size - 1; i >= 0; --i) {
                final InstructionContextImpl current = this.executionPredecessors.get(i);
                final Instruction currentlast = current.getInstruction().getInstruction();
                if (currentlast instanceof RET) {
                    ++retcount;
                }
                if (currentlast instanceof JsrInstruction && --retcount == -1) {
                    return current;
                }
            }
            return null;
        }
        
        private boolean mergeInFrames(final Frame inFrame) {
            final Frame inF = this.inFrames.get(this.lastExecutionJSR());
            final OperandStack oldstack = inF.getStack().getClone();
            final LocalVariables oldlocals = inF.getLocals().getClone();
            try {
                inF.getStack().merge(inFrame.getStack());
                inF.getLocals().merge(inFrame.getLocals());
            }
            catch (final StructuralCodeConstraintException sce) {
                this.extendMessageWithFlow(sce);
                throw sce;
            }
            return !oldstack.equals(inF.getStack()) || !oldlocals.equals(inF.getLocals());
        }
        
        @Override
        public void setTag(final int tag) {
            this.TAG = tag;
        }
        
        @Override
        public String toString() {
            return this.getInstruction().toString(false) + "\t[InstructionContext]";
        }
    }
}
